package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.ctre.phoenix6.hardware.CANcoder;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkBase.PersistMode;
import com.revrobotics.spark.SparkBase.ResetMode;
import com.revrobotics.spark.SparkLowLevel.MotorType;

import edu.wpi.first.math.MathUtil;
import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.math.util.Units;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ManipulatorPivotConstants;

public class ManipulatorPivot extends SubsystemBase {
    protected SparkMax armMotor;

    private CANcoder canCoder;

    private PIDController positionController;
    private PIDController velocityController;

    private ArmFeedforward feedForward;

    public ManipulatorPivot() {
        armMotor = new SparkMax(
            ManipulatorPivotConstants.kArmMotorID, 
            MotorType.kBrushless
        );

        armMotor.configure(
            ManipulatorPivotConstants.motorConfig, 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        positionController = new PIDController(
            ManipulatorPivotConstants.kPositionalP, 
            ManipulatorPivotConstants.kPositionalI, 
            ManipulatorPivotConstants.kPositionalD
        );
        
        // TODO: Generate constants for continuous input range based on CANcoder configuration?
        positionController.enableContinuousInput(Units.degreesToRadians(-180), Units.degreesToRadians(179));
        positionController.setTolerance(ManipulatorPivotConstants.kPositionalTolerance);

        velocityController = new PIDController(
            ManipulatorPivotConstants.kVelocityP, 
            ManipulatorPivotConstants.kVelocityI, 
            ManipulatorPivotConstants.kVelocityD
        );

        velocityController.setTolerance(ManipulatorPivotConstants.kVelocityTolerance);

        canCoder = new CANcoder(ManipulatorPivotConstants.kCANcoderID);
        canCoder.getConfigurator().apply(ManipulatorPivotConstants.canCoderConfig);
    }

    /**
     * Returns whether or not the motion is safe relative to the encoder's current position
     * and the arm safe stow position
     * 
     * @return Is the motion safe
     */
    public boolean isMotionSafe() {
        return isMotionSafe(getEncoderPosition());
    }

    /**
     * Returns whether or not the motion is safe relative to some target position and the
     * arm safe stow position
     * 
     * @param motionTarget The target position to determine the safety of
     * @return Is the motion safe
     */
    public boolean isMotionSafe(double motionTarget) {
        return motionTarget > ManipulatorPivotConstants.kArmSafeStowPosition;
    }

    /**
     * A manual rotation command that will move the elevator using a consistent velocity disregarding direction
     * 
     * @param speed The velocity at which the arm rotates
     * @return Sets motor voltage to achieve the target velocity
     */
    public Command runAssistedPivot(DoubleSupplier speed) {
        double clampedSpeed = MathUtil.clamp(
            speed.getAsDouble(), 
            -1, 
            1
        );

        return run(() -> {
            double realSpeedTarget = clampedSpeed * ManipulatorPivotConstants.kPivotMaxVelocity;

            double voltsOut = velocityController.calculate(
                getEncoderVelocity(), 
                realSpeedTarget
            ) + feedForward.calculate(
                getEncoderPosition(), 
                getEncoderVelocity()
            );

            armMotor.setVoltage(voltsOut);
        });
    }

    /**
     * Moves the arm to a target destination (setpoint)
     * 
     * @param setpoint Target destination of the subsystem
     * @param timeout Time to achieve the setpoint before quitting
     * @return Sets motor voltage to achieve the target destination
     */
    public Command goToSetpoint(double setpoint, double timeout) {
        double clampedSetpoint = MathUtil.clamp(
            setpoint, 
            0, 
            ManipulatorPivotConstants.kRotationLimit
        );

        return run(() -> {
            double voltsOut = positionController.calculate(
                getEncoderPosition(), 
                clampedSetpoint
            ) + feedForward.calculate(
                getEncoderPosition(), 
                getEncoderVelocity()
            );

            armMotor.setVoltage(voltsOut);
        }).until(positionController::atSetpoint).withTimeout(timeout);
    }

    /**
     * Returns the CANCoder's position in radians
     * 
     * @return CANCoder's position in radians
     */
    public double getEncoderPosition() {
        return Units.rotationsToRadians(canCoder.getAbsolutePosition().getValueAsDouble());
    }

    /**
     * Returns the CANCoder's velocity in radians per second
     * 
     * @return CANCoder's velocity in radians per second
     */
    public double getEncoderVelocity() {
        return Units.rotationsToRadians(canCoder.getVelocity().getValueAsDouble());
    }
}
